From a169f4261024397dd3ddb944decc1601a623df2a Mon Sep 17 00:00:00 2001
From: Neha Malcom Francis <n-francis@ti.com>
Date: Mon, 25 Jul 2022 18:28:09 -0500
Subject: [PATCH] HACK: j721s2: QoS workaround for GPU cache incoherency

GPU <-> CPU memory inconsistencies have been found in BXS and J721s2.
Every once in a while, the GPU's view of the physical memory differs
from the CPU's view of the physical memory. In such cases, the data
written by CPU to be read by GPU, is not seen by the GPU. This needs
a fix in the Linux GPU driver.

The temporary workaround for now is:
- Set GPU aType to 3 (bypass MSMC coherency)
- Have L3 cache disabled (default in SDK)

This is a temporary fix that will be removed after the 8.4 release.
The GPU driver will then been rewritten to not access memory with
mismatched attributes. Until then undefined behavior will occur
and the following should be avoided:
 * Passing the GPU cached buffers (no EXT_image_dma_buf_import)
 * Using L3 cache (do not enable it in board config)
 * Reuse memory used by the GPU

Signed-off-by: Neha Malcom Francis <n-francis@ti.com>
[afd: Rebased on ti-u-boot-2021.01, fixed build warning, fixed commit message]
Signed-off-by: Andrew Davis <afd@ti.com>
---
 arch/arm/mach-k3/j721s2_init.c | 20 ++++++++++++++++++++
 1 file changed, 20 insertions(+)

diff --git a/arch/arm/mach-k3/j721s2_init.c b/arch/arm/mach-k3/j721s2_init.c
index 38693a6546..2a345fd6b7 100644
--- a/arch/arm/mach-k3/j721s2_init.c
+++ b/arch/arm/mach-k3/j721s2_init.c
@@ -22,6 +22,10 @@
 #include <mmc.h>
 #include <remoteproc.h>
 
+#define GPU_HARDWARE_BASE_ADDRESS_READ  0x45DC5100
+#define GPU_HARDWARE_BASE_ADDRESS_WRITE 0x45DC5900
+#define REGISTER_VALUE_DISABLE_SNOOP    0x30000000
+
 #ifdef CONFIG_SPL_BUILD
 
 #ifdef CONFIG_TI_SECURE_DEVICE
@@ -99,6 +103,20 @@ static void ctrl_mmr_unlock(void)
 	mmr_unlock(CTRL_MMR0_BASE, 7);
 }
 
+static inline void _qos_work_around_for_gpu(phys_addr_t base)
+{
+	u32 offset;
+
+	for (offset = 0; offset < 160; offset++)
+		writel(REGISTER_VALUE_DISABLE_SNOOP, base + (offset * 4));
+}
+
+static void qos_work_around_for_gpu(void)
+{
+	_qos_work_around_for_gpu(GPU_HARDWARE_BASE_ADDRESS_READ);
+	_qos_work_around_for_gpu(GPU_HARDWARE_BASE_ADDRESS_WRITE);
+}
+
 void k3_mmc_stop_clock(void)
 {
 	if (IS_ENABLED(CONFIG_K3_LOAD_SYSFW)) {
@@ -159,6 +177,8 @@ void board_init_f(ulong dummy)
 	if (IS_ENABLED(CONFIG_CPU_V7R)) {
 		disable_linefill_optimization();
 		setup_k3_mpu_regions();
+	} else {
+		qos_work_around_for_gpu();
 	}
 
 	/* Init DM early */
-- 
2.34.1

